\ 3iForth.fs -- 3-Instruction Forth for AVR Atmel 328 / Arduino Nano
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ 2020-06-22

\ See Article & Discussion at http://www.mathscitech.org/articles/3iForth
\ TARGET = 3iForth.avr running on ATmel328 in Arduino Nano (compiled & uploaded to Nano using free Arduino IDE & default bootloader)
\ HOST = 3iForth.fs running in GForth 0.7.0 on Windows PC
\ Nano on startup (or reset) sets up serial port (9600 baud), initializes data stack, and waits for instructions over serial line

\ c. 250 lines of Forth code

\ USAGE:  
\   1. Connect Arduino Nano to HOST using USB mini cable
\           PINS - note: A0 through A7 are pins 14 to 21.  A0 to A5 (14-19) are I/O, A6/A7 (20-21) are AnalogRead only.
\   2. Run RealTerm with command line setup: baud=9600 half port=6.  This sets up the serial ports correctly to use 
\   3. HOST > run Gforth > include 3iForth.fs
\   4. HOST > help, help1, help2   (provides instructions)
\ Demo board 1: LED board with 11 LEDs alternating red and yellow (pins 2-12) and a tiny speaker (pin 13)
\ Demo board 2: Programmable Digital Synthesizer board with 13 keys and speaker

( * SERIAL PORT - these words open and close a given COM port on the PC * )
(   ex:            5 open-com              )
(   MISSING:  need to add GForth code for configuring the COM port.  Workaround: Use RealTerm to connect first  )

decimal
$1B constant ESC
: not ( 0/x -- 1/0 ) if 0 else 1 then ;
: not4 ( $xx -- $yy ) invert $F and ; \ 4-bit complement
: not8 ( $xx -- $yy ) invert $FF and ; \ 4-bit complement


variable COMhandle
    : save-handle ( handle -- ) COMhandle ! ;
    : fh ( -- handle ) COMhandle @ ;

: com-string ( port# -- c-adr cnt )
    \ Builds COM port string.  Test: 6 com-string type
    >r s" comX" 2dup + 1 - '0' r> + swap c! ; 
: com-open? ( handle errcode -- )
    dup 0= ." ..." 
    if drop save-handle ." COM port opened ... " 
    else cr ." **error: check device is plugged into correct COM port" cr throw
    then ;
:  [OPEN-COM] ( port# --  )  \ Opens COM port in binary mode
    com-string r/w bin create-file com-open? 1500 ms ." OK" ;  \ Settling time delay of 1000ms doesn't work, 1500ms works
:  CLOSE-COM ( -- )
    fh close-file throw cr ." COM port closed." cr ;


( * MEMORY  - these words are essential for working efficiently with memory buffers, i.e. arrays of memory * )

: bytes  1 * ;
: wwords 2 * ;   
: dwords 4 * ;   \ GForth is 32-bits
\ Efficient fetch & store from/into and an array
\ BYTES
:  c@+ ( adr -- adr' val )  
    dup 1 bytes + swap    c@ ;
:  c@- ( adr -- adr' val )  
    dup 1 bytes - swap    c@ ;
:  c!+ ( val adr -- adr' )  
    dup 1 bytes + -rot    c! ;
\ WORDS
: bytes2word ( hi8 lo8 -- val16 )
	swap    8 lshift   or ;
: lohi ( val16 -- hi8 lo8  )  \ little endian - low on top of stack.  ex: $FFFE lohi yields $FF $FE
    dup   $FF00 and   8 rshift    swap $FF and ;
    \ 440 f2hl 1 184
: ww@ ( adr -- val16 )  \ little endian retrieval
	c@+    swap c@    swap  bytes2word ;
: ww@+  ( adr -- adr' val16 )  
	c@+    swap c@+   swap -rot  swap  bytes2word ;
: ww!  ( val16 adr -- )  \ little endian storage
	swap lohi rot  c!+ c! ;
: ww!+ ( val16 adr -- adr' )  
	swap lohi rot  c!+ c!+ ;
\ DWORDS
: @+ 
    dup 1 dwords + swap @ ;
: !+
    dup 1 dwords + -rot ! ;
: 4drop 2drop 2drop ;  \ for 4-bit port work


( * SERIAL READ - get messages from target into receive buffer * )

create RCV-BUFFER 255 bytes allot  \ buffer for holding messages from the uC
: show-rcv ( cnt -- cnt )   \  ex: 255 show-rcv
    \ Prints rcv buffer in decimal  ( use hex. instead of . to show values in hex  )
    dup RCV-BUFFER swap dup 0> 
    if 0 cr ." <rcv: " do c@+ . ." " loop else drop then drop ;    
: read-com ( n -- cnt )   \ read N bytes  ex: 50 read-com show-rcv
    RCV-BUFFER swap fh read-file throw ( -- cnt ) ;
\ : bytes. ( cnt -- cnt ) dup ." >> " . ." bytes on port" ; 

\ COMPLICATION: for REAL-TIME read, i.e. after sending a serial command and expecting a response from the uC on Serial.  80 ms works, 60 intermittent, 25 doesn't work
: rxN ( N -- cnt ) 80 ms read-com ( bytes. ) show-rcv ;     
: rx ( -- ) 50 rxN drop ;

( * SERIAL TRANSMIT - send messages in prepared message buffer to target * )

create MSG-BUFFER 20 bytes allot     \  write buffer
: show-msg ( cnt -- )                \ ex: 10 show-msg
    MSG-BUFFER swap 0 cr ." MSG-BUFFER>: " do c@+ . ." " loop drop ;          \ change . to hex. to show hex values
: [tx] ( msg-buf len -- )  \ serial transmit len bytes from msg-buf
    fh write-file throw fh flush-file throw  ; 
: tx  ( adr'  -- ) MSG-BUFFER swap over -  [tx] ;
: tx1 ( byte -- ) pad c!  pad 1 [tx] ;
: tx2 ( 2nd 1st -- ) pad c!+ c!   pad 2 [tx] ;  \ serial transmit double byte, needed for xstore

( * SENDING REMOTE INSTRUCTIONS TO TARGET - some work with parameters that must reside on the target's data-stack * )

\ Notes: 
\ 1.   tx ;      only xcall eventually transits.  the others prepare the MSG-BUFFER (fewer comms -> more efficient)
\ 2.   -- adr'   each function accepts an address and provides the next address for easy chaining into MSG-BUFFER

\ These are 3iForth commands, as distinct from exported chip instructions
0 constant CMD_X_S  \ maps to stack_len() in 3iForth.avr
1 constant CMD_XSTORE
2 constant CMD_XDROP
3 constant CMD_XFETCH
4 constant CMD_XCALL

\ Note: stack diagram is the same for .fs and for .avr.  This is because the [store] prep function reverses the entries
\ CODE REVIEW: CHANGE TO USE REVERSE STACK COMMAND using roll
: [newmsg] ( -- adr' ) MSG-BUFFER ;
: [store]  ( a adr -- adr' ) CMD_XSTORE swap c!+ c!+ ;
: [store2] ( a b adr -- adr' )  >r swap r> [store] [store] ;
: [store3] ( a b c adr -- adr' ) swap >r [store2] r> swap [store] ;
: [store4] ( a b c d adr -- adr' ) swap >r [store3] r> swap [store] ;
: [store5] ( a b c d e adr -- adr' ) swap >r [store4] r> swap [store] ; 

\ remote stack allows testing the chip remotely using the local stack as if it were a local development, and without risking pins etc
: xfetch  ( -- val )  CMD_XFETCH tx1 80 ms 1 read-com drop rcv-buffer c@ ;  \ fetch top byte from remote stack & return to TOS local
: x_s     ( -- )      cr CMD_X_S tx1 rx xfetch ." (". ." )" ;  \ queries chip and reports size of stack in bytes
: xstore  ( a -- )    CMD_XSTORE tx2 ;          \ store byte a onto remote stack
: xdrop   ( -- )      CMD_XDROP tx1 ;   \ remove a byte from remote stack
: 2xdrop  xdrop xdrop ;
: 4xdrop  2xdrop 2xdrop ;
: xcall   ( instr -- ) xstore CMD_XCALL tx1 ;     \ call export chip instruction (see .avr fcall jump table)

: open-com ( usb# -- ) [open-com] 5 rxN drop xdrop ;

( * HOST INTERFACE  CALL - on the host machine we don't need to know the mechanism, we think we live in the target and call the exported functions directly using Forth * )

\ WARNING: these match the jumptable entries in the chip
0 constant  CALL_PINMODE
1 constant  CALL_PINWRITE
2 constant  CALL_DELAY
3 constant  CALL_TONE
4 constant  CALL_SETREG
5 constant  CALL_SWITCHKEY  \ digital piano only
6 constant  CALL_RECORD     \ digital piano only
7 constant  CALL_PINREAD
8 constant  CALL_KEYPAD     \ keypad only
9 constant  CALL_PINREADLEVEL
$B constant CALL_PORT4MODE
$C constant CALL_PORT4WRITE
$D constant CALL_PORT4READ
$E constant CALL_CHECKREG

1 constant OUTPUT
0 constant INPUT
2 constant INPUT_PULLUP
\ NO built-in Arduina constant for INPUT_PULLDOWN

: pinMode ( mode pin -- )  swap xstore xstore CALL_PINMODE xcall 60 ms ;   
    \ ex: 1 9 pinmode  (output)  modes: 0x0 INPUT, 0x1 OUTPUT, 0x2 INPUT_PULLUP
    \ Note: need settle time on pin when changing mode.  60 ms is magic number
: out ( pin -- )   OUTPUT swap pinmode ;
: in ( pin -- )    INPUT swap pinmode ;
: pullup ( pin -- ) INPUT_PULLUP swap pinmode ;  \ ex: 2 pullup
: pullup-keys  \ set all keyboard pins 2-12 to INPUT_PULLUP
    13 2 do i pullup loop 
    \ NEED TO ALSO DO A0 and A1 -- not sure how to pass those parameters!
    100 ms ; 

1 constant HIGH_SIGNAL
0 constant LOW_SIGNAL

: pinwrite ( level pin -- )  swap xstore xstore CALL_PINWRITE xcall ;  \ ex: 1 9 pinwrite   0 9 pinwrite
: high ( pin -- )  HIGH_SIGNAL swap pinwrite ;
: low ( pin -- )   LOW_SIGNAL swap pinwrite ;
: lo low ;
: hi high ;
: testpin ( n -- )  dup dup out high 500 ms low ." done." ;

\ FIX NEEDED: ms is usually more than 1 byte worth
: delay xstore CALL_DELAY xcall ; \ ex: 128 delay   NEEDED on TARGET to ensure timings are preserved when running untethered

: [tone] ( freq-hi freq-lo dur-hi dur-lo pin -- ) [newmsg] [store5] tx CALL_TONE xcall ;
   \ 13 setspeaker
   \ 440 lohi 500 lohi speaker [tone] 600 ms 440 lohi 500 lohi speaker [tone]  \ ringing tone! :)
   \ https://www.arduino.cc/reference/en/language/functions/advanced-io/tone/
   \ https://github.com/bhagman/Tone#ugly-details

: setreg ( val16 reg -- )   \ ex: 440 0 setreg    sets register0 to 16-bit value   ( e.g. musical note A4=440 Hz )
    swap lohi rot [newmsg] [store3] tx CALL_SETREG xcall ;
: checkreg ( reg -- serial: hi lo ) xstore CALL_CHECKREG xcall rx ;

: switchkey ( -- )  CALL_SWITCHKEY xcall ;  \ switch hardware modes
: record  ( on1/off0 -- ) xstore CALL_RECORD xcall ; \ turn on/off recording
	
: pinread ( pin -- 0/1 )   \ digitalRead() - with realtime delay (COMPLICATION)  -- WARNING: set pin mode to INPUT (active high) or PULLUP (active low)
	xstore CALL_PINREAD xcall
	100 ms   \ 50 ms seems to work ( 0 failures out of 30 trials ), if processor is also writing to display, then 100ms is enough
	1 read-com if rcv-buffer c@ else cr ." FAIL pinread: no bytes" cr then ;
: poll-pin ( pin -- ) begin dup pinread . key? until ." Done!" drop ;   \ for active low.  if active high remove the not

: pinreadLevel ( Apin -- counts )   \ analogRead() -- NOTE: use ANALOG pin (A0-A7, 14-21) - result in 10 bit counts (0 to 1024)
    \ WARNING: set pin mode to INPUT
    \ Ref: https://www.arduino.cc/reference/en/language/functions/analog-io/analogread/
	xstore CALL_PINREADLEVEL xcall
	60 ms   \ 50 ms seems to work ( 0 failures out of 30 trials )
	2 read-com if rcv-buffer c@+ 8 lshift swap c@ + else cr ." FAIL pinread: no bytes" cr then ;
: Vref5 ( cnts -- mV ) 49 10 */ ;  \ 4.9mV per count  (5V=1024 counts)
: poll-level ( Apin -- ) begin dup pinreadLevel Vref5 . cr 100 ms key? until ." Done!" drop ;
: plot-level ( Apin -- ) begin dup pinreadLevel Vref5 dup 100 / cr spaces ." XX  " . 100 ms key? until ." Done!" drop ;


: pollkeypad ( -- val )  \ read matrix membrane keypad and send back 0x0-0xF numbers
	CALL_KEYPAD xcall
	200 ms
	1 read-com if rcv-buffer c@ else $FF then ;

( * setting up and reading ports --- IN PROGRESS 2020-09-22---Tue---07:52 * )


\  4-bit port setup
15 constant REG_PORTOUT
16 constant REG_PORTIN
\ Physical Port Control   Ex:  5 4 3 2  pins for output port, lsb TOS.   Ex: 11 10 9 8 pins for input port, lsb TOS
: join4 ( msb c b lsb -- mcbl-16bits )  \  ex: 5 4 3 2 -- $5432 (in hex) combines 4 bytes (pins) to 16-bit value for setreg
    swap 4 lshift or
    swap 8 lshift or
    swap $C lshift or ;

: port4out ( d c b a -- )  \ ex: 5 4 3 2  port4out, lsb TOS      Test: reg_portout checkreg
    join4 REG_PORTOUT setreg 80 ms
    OUTPUT xstore REG_PORTOUT xstore CALL_PORT4MODE xcall ;
: port4in  ( d c b a -- )  \ ex: 11 10 9 8  port4in, lsb TOS     Test: reg_portin checkreg
    join4  REG_PORTIN  setreg 80 ms
    INPUT xstore  REG_PORTIN xstore CALL_PORT4MODE xcall ;

: port4write ( hex -- ) xstore REG_PORTOUT xstore CALL_PORT4WRITE xcall ;           \ writes hex digit to 4-bit port
: port4read ( -- hex )         \ reads hex digit from 4-bit port
    REG_PORTIN  xstore CALL_PORT4READ  xcall 
    60 ms
    1 read-com if rcv-buffer c@ else cr ." FAIL port4read: no bytes" cr then ;
: poll-port ( -- ) begin port4read not4 dup if . cr else ." ." drop then key? until ." Done!" ;  \ use not4 for active low


( * PROBLEM ORIENTED LANGUAGE - ON THE HOST --- setting up a simplified yet powerful custom dictionary * )

: reload ( -- )   \ ensure com port is closed before reloading which clobbers fh so need to open-com port again
  fh 0> if close-com then
  s" 3iForth.fs" included ;

( * HELP FILES * )

: help
    cr cr cr cr cr
    cr ."  ------------------------------------------------------------"
    cr ."  3-instruction Forth for Atmel 328 chip:"
    cr ."     STARTING/STOPPING"
    cr ."       5 open-com    -  opens communication to port 5 "
    cr ."       close-com "
    cr ."       reload "
    cr ."     OUTPUT PINS"
	cr ."       2 out         - sets pin 2 into pullup high mode"
    cr ."       2 high        - sends voltage to pin 2  (choose pins 2 through 12)"
    cr ."       2 low         - sends voltage to pin 2  (choose pins 2 through 12)"
	cr ."	 INPUT PINS"
    cr ."       2 pullup     - sets pin 2 into pullup high mode"
	cr ."       2 pinread
	cr ."    LOWEST LEVEL => help1"
    cr ."       host:     x_s   1 xstore   xdrop  xfetch .   <instr> xcall   "
    cr ."       target:   ms  pinmode   pinwrite   pinread "
    cr ."                 pinwriteLevel pinreadLevel  setreg  checkreg  "
    cr ."                 port4mode  port4write  port4read  tone  "
    cr ."    Add-ons:"
    cr ."    include   3iLight.fs   "
    cr ."    include   3iPort-4x4.fs 3iSound-4x4.fs  3iPlaylist.fs  3iRecord.fs  3iGinan.fs  "
    cr ."    include   3iSensor.fs  
    cr cr
;

: help1 ( -- )
    cr ."  2.  1  13 pinmode       In general:  pinMode ( mode pin --  )"
    cr ."  2.  2   9 pinmode       set pin 9 to INPUT_PULLUP
    cr ."  3.  1  13 pinwrite        In general:  pinwrite  ( level pin -- )   (called digitalWrite in Arduino C)."
    cr ."  4.  0  13 pinwrite        "
;

\ ======================
help


\    cr
\    cr ."   => LEDs & BUZZERS:  include 3iLight.fs"
\    cr ."       all-out  all-high all-low  5 n-vivid
\    cr ."   => SOUND / MUSIC (tone)   include 3iSound.fs"  
\    cr ."       start-speaker  - sets pin 13 to speaker" 
